"""

(c)2012 Martin Tarenskeen <m.tarenskeenATzonnetDOTnl>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import array
import os
import zipfile
from tempfile import mkstemp

from . import dx7
from . import dx200
from . import syxmidi
from . import fourop
from . import tx7
from . import fb01
from . import pssx80
from . import vopm
from . import dxcommon

MAXFILESIZE=dxcommon.MAXFILESIZE
PROGRAMNAME="DXconvert"
PROGRAMVERSION=dxcommon.PROGRAMVERSION
PROGRAMDATE=dxcommon.PROGRAMDATE
ALLOWED_CHARACTERS=dxcommon.ALLOWED_CHARACTERS
MID_IN=dxcommon.MID_IN
MID_OUT=dxcommon.MID_OUT

########## FILE INPORT/EXPORT #########

def read(infile, offset=0, check=False, mid_in=MID_IN, mid_out=MID_OUT):
    dx7data = []
    dx72data = []
    tx7data = []
    channel = 0
    if zipfile.is_zipfile(infile):
        with zipfile.ZipFile(infile, 'r') as zf:
            zflist = zf.namelist()
            for n in zflist:
                try:
                    dx7dat = []
                    dx72dat = []
                    tx7dat = []
                    d = zf.read(n)
                    tmp = mkstemp()[1]
                    with open(tmp, 'wb') as f:
                        f.write(d)
                    dx7dat, dx72dat, tx7dat, channel = read2(mid_in, mid_out, tmp, offset, check)
                    os.remove(tmp)
                    dx7data += dx7dat
                    dx72data += dx72dat
                    tx7data += tx7dat
                    print "\t{}".format(n)
                except:
                    print "WARNING: only {} from {} files in zip were read".format(zflist.index(n)+1, len(zflist))
                    break
    else:
        dx7data, dx72data, tx7data, channel = read2(mid_in, mid_out, infile, offset, check)
    return dx7data, dx72data, tx7data, channel

def read2(mid_in, mid_out, infile, offset=0, check=False):
    ext = os.path.splitext(infile)[1].lower()
    if MAXFILESIZE == 0:
        size = os.path.getsize(infile)-offset
    else:
        size = min(MAXFILESIZE, os.path.getsize(infile))-offset
    data = array.array("B")
    with open(infile, 'rb') as f:
        data.fromfile(f, offset+size)
    data = data.tolist()[offset:]

    #Dump Request
    if dxcommon.ENABLE_MIDI:
        if ext==".req" and data[:2] == [0xf0, 0x43]:
            data, size = dxcommon.req2data(data, mid_in, mid_out)


    data += [0xf7, 0xf7, 0, 0, 0, 0, 0, 0, 0, 0]
    dx7data = []
    dx72data = []
    tx7data = []
    channel = 0

    #check if file is a MIDI file
    midskip=0
    if data[0:4] == [0x4d, 0x54, 0x68, 0x64]: #"MThd"
        midskip=2

    #single voice(s)
    if size>160:
        for i in range(size-51):
            if data[i]==0xf0 and data[i+1+midskip//2]==0x43 and data[i+2+midskip//2]<=0x0f and data[i+3+midskip//2]==5 and data[i+4+midskip//2]==0 and data[i+5+midskip//2]==0x31 and data[i+6+50+midskip//2]==0xf7:
                if (check==False) or (dxcommon.checksum(data[i+6+midskip//2:i+6+midskip//2+49]) == data[i+6+midskip//2+49]):
                    dx72data += dx7.aced2amem(data[i+6+midskip//2:i+6+midskip//2+49])
                else:
                    dx72data += dx7.initamem()
                channel=data[i+2+midskip//2]
        for i in range(size-157):
            if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0 and data[i+4+midskip]==1 and data[i+5+midskip]==0x1b and data[i+6+156+midskip]==0xf7:
                if (check==False) or (dxcommon.checksum(data[i+6+midskip:i+6+midskip+155])==data[i+6+midskip+155]):
                    dx7data += dx7.vced2vmem(data[i+6+midskip:i+6+midskip+155])
                else:
                    dx7data += dx7.initvmem()
                channel=data[i+2+midskip]

    #packed 32 voices
    if size>4101:
        for i in range(size-4098):
            #data[i+4+midskip] should be 0x20 but often is 0x10 
            if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0x09 and (data[i+4+midskip]==0x10 or data[i+4+midskip]==0x20) and data[i+5+midskip]==0x00 and data[i+6+4097+midskip]==0xf7:
                if (check==False) or (dxcommon.checksum(data[i+midskip+6:i+midskip+6+4096]) == data[i+midskip+6+4096]):
                    dx7data += data[i+midskip+6:i+midskip+6+4096]
                else:
                    dx7data += dx7.initvmem()*32
                channel = data[i+2+midskip]
        for i in range(size-1122):
            if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0x06 and data[i+4+midskip]==0x08 and data[i+5+midskip]==0x60 and data[i+6+1121+midskip]==0xf7:
                if (check==False) or (dxcommon.checksum(data[i+midskip+6:i+midskip+6+1120]) == data[i+midskip+6+1120]): 
                    dx72data += data[i+midskip+6:i+midskip+6+1120]
                else:
                    dx72data += dx7.initamem()*32
                channel=data[i+2+midskip]

    # Yamaha DX21/TX81Z VMEM
    for i in range(size-4098):
        if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0x04 and data[i+4+midskip]==0x20 and data[i+5+midskip]==0x00 and data[i+6+4097+midskip]==0xf7:
            print "WARNING: 4op to 6op conversion"
            if (check==False) or (dxcommon.checksum(data[i+6+midskip:i+6+4096+midskip])==data[i+6+4096+midskip]):
                for p in range(32):
                    dx7data += fourop.vmm2vmem(data[128*p+i+6+midskip:128*(p+1)+i+6+midskip])[0]
                    dx72data += fourop.vmm2vmem(data[128*p+i+6+midskip:128*(p+1)+i+6+midskip])[1]
            else:
                dx7data += dx7.initvmem()*32
                dx72data += dx7.initamem()*32

    vcd = []
    for i in range(size-95):
        if data[i]==0xf0 and data[i+1+midskip//2]==0x43 and data[i+2+midskip//2]<=0x0f and data[i+3+midskip//2]==0x03 and data[i+4+midskip//2]==0x00 and data[i+5+midskip//2]==0x5d:
            #VCED
            vcd += data[i+6+midskip//2:i+6+midskip//2+93]
    
    acd2 = []
    for i in range(size-12):    
        if data[i]==0xf0 and data[i+1+midskip//2]==0x43 and data[i+2+midskip//2]<=0x0f and data[i+3+midskip//2]==0x7e and data[i+4+midskip//2]==0x00 and data[i+5+midskip//2]==0x14 and dxcommon.list2string(data[i+6+midskip//2:i+16+midskip//2])=='LM__8023AE':
            #ACED2
            acd2 += data[i+16+midskip//2:i+16+midskip//2+10]
    
    acd = []
    for i in range(size-25):    
        if data[i]==0xf0 and data[i+1+midskip//2]==0x43 and data[i+2+midskip//2]<=0x0f and data[i+3+midskip//2]==0x7e and data[i+4+midskip//2]==0x00 and data[i+5+midskip//2]==0x21 and dxcommon.list2string(data[i+6+midskip//2:i+16+midskip//2])=='LM__8976AE':
            #ACED
            acd += data[i+16+midskip//2:i+16+midskip//2+23]
    
    count = len(vcd)//93
    delay = fourop.initdelay()*count
    efeds = fourop.initefeds()*count
    acd3 = fourop.initacd3()*count
    if len(acd2) != 10*count: acd2 = fourop.initacd2()*count
    if len(acd) != 23*count: acd = fourop.initacd()*count

    for v in range(count):
        vmm = fourop.vcd2vmm(
            vcd[93*v:93*(v+1)], 
            acd[23*v:23*(v+1)], 
            acd2[10*v:10*(v+1)], 
            acd3[20*v:20*(v+1)], 
            efeds[6*v:6*(v+1)], 
            delay[2*v:2*(v+1)]
            )
        dx7data += fourop.vmm2vmem(vmm)[0]
        dx72data += fourop.vmm2vmem(vmm)[1]

    # Yamaha FB01 RAM and ROM voicebanks
    for i in range(size-6360):
        if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]==0x75 and data[i+3+midskip]<=0x0f and data[i+6362+midskip]==0xf7:
            print "WARNING: FB01 to 6op conversion"
            for p in range(48):
                fb=[]
                if (check==False) or (dxcommon.checksum(data[i+midskip+76+131*p:i+midskip+76+131*p+128])==data[i+midskip+76+131*p+128]):
                    for n in range(64):
                        fb.append(data[i+midskip+76+131*p+2*n] + (data[i+midskip+76+131*p+2*n+1]<<4))
                    dx7data += fb01.fb2vmem(fb)[0] 
                    dx72data += fb01.fb2vmem(fb)[1]  
                else:
                    dx7data += dx7.initvmem()
                    dx72data += dx7.initamem()

    # Yamaha FB01 voice bank 0
    for i in range(size-6357):
        if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0x0c and data[i+6359+midskip]==0xf7:
            print "WARNING: FB01 to 6op conversion" 
            for p in range(48):
                fb=[]
                if (check==False) or (dxcommon.checksum(data[i+midskip+73+131*p:i+midskip+73+131*p+128])==data[i+midskip+73+131*p+128]):
                    for n in range(64):
                        fb.append(data[i+midskip+73+131*p+2*n] + (data[i+midskip+73+131*p+2*n+1]<<4))
                    dx7data += fb01.fb2vmem(fb)[0]
                    dx72data += fb01.fb2vmem(fb)[1]
                else:
                    dx7data += dx7.initvmem()
                    dx72data += dx7.initamem()

    # Yamaha FB01 instrument 1~8
    for i in range(size-136):
        if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]==0x75 and data[i+3+midskip]<=0x0f and (data[i+4+midskip]&8) and data[i+138+midskip]==0xf7:
            print "WARNING: FB01 to 6op conversion"
            fb=[]
            if (check==False) or (dxcommon.checksum(data[i+midskip+9:i+midskip+9+128])==data[i+midskip+9+128]):
                for n in range(64):
                    fb.append(data[i+midskip+9+2*n] + (data[i+midskip+9+2*n+1]<<4))
                dx7data += fb01.fb2vmem(fb)[0]
                dx72data += fb01.fb2vmem(fb)[1]
            else:
                dx7data += dx7.initvmem()
                dx72data += dx7.initamem()


    # TX7 1 Performance
    for i in range(size-96):
        if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0x01 and data[i+4+midskip]==0x00 and data[i+5+midskip]==0x5e:
            if (check==False) or (dxcommon.checksum(data[i+6+midskip:i+6+midskip+94])==data[i+6+midskip+94]):
                tx7data += tx7.pced2pmem(data[i+6+midskip:i+6+midskip+94])
            else:
                tx7data += tx7.initpmem()

    # TX7 32+32 Performances
    for i in range(size-4098):
        if data[i]==0xf0 and data[i+1+midskip]==0x43 and data[i+2+midskip]<=0x0f and data[i+3+midskip]==0x02 and data[i+4+midskip]==0x20 and data[i+5+midskip]==0x00:
            if (check==False) or (dxcommon.checksum(data[i+6+midskip:i+6+midskip+4096])==data[i+6+midskip+4096]):
                tx7data += data[i+6+midskip:i+6+midskip+2048]
            else:
                tx7data += tx7.initpmem()
    
    # PSS 480/580/680
    for i in range(size-68):
        if data[i]==0xf0 and data[i+1+midskip//2]==0x43 and data[i+2+midskip//2]==0x76 and data[i+3+midskip//2]==0x00 and data[i+midskip//2+71]==0xf7:
            if (check==False) or (dxcommon.checksum(data[i+4+midskip//2:i+4+midskip//2+66])==data[i+4+midskip//2+66]):
                print "WARNING: experimental PSS-480/580/680 2-OP to 6-OP conversion"
                pss=[]    
                for n in range(33):
                    pss.append((data[i+midskip//2+4+2*n]<<4) + data[i+midskip+4+2*n+1])
                dx7data += pssx80.pss2vmem(pss, infile)
            else:
                dx7data += dx7.initvmem()

    # Still not found anything? Then keep searching ...
    if dx7data != []:
        pass
    
    #Dr.T TX7 files, Steinberg
    elif (ext in (".snd", ".tx7")) and size==8192:
        dx7data += data[:4096]
        tx7data += data[4096:4096+2048]

    #Transform XSyn
    elif ext==".bnk" and size==8192:
        for i in range(32):
            dx7data += data[256*i:256*i+128]
            tx7data += data[256*i+128:256*i+192]

    #DX200 DX2
    elif ext==".dx2" and size==326454:
        for patch in range(128):
            dx7data += dx200.dx2tovmem(data[34+patch*381:34+(patch+1)*381])
            dx72data += dx200.dx2toamem(data[34+patch*381:34+(patch+1)*381])

    #Yamaha DX Simulator DXC
    elif ext==".dxc" and size==18004:
        for patch in range(64):
            dx7data += dx200.dx2tovmem(data[20+patch*281:20+(patch+1)*281])
            dx72data += dx200.dx2toamem(data[20+patch*281:20+(patch+1)*281])

    #Voyetra SIDEMAN and PATCHMASTER
    elif data[0:4] == [0xdf, 0x05, 0x01, 0x00]:
        if ext==".b13" and size==5663:
            if dxcommon.list2string(data[6:19]) == "YAMAHA DX7/TX":
                dx7data += data[0x060f:0x160f]

        if ext == ".b15" and size==10755:
            if dxcommon.list2string(data[6:18]) == "YAMAHA TX81Z":
                for i in range(32):
                    dx7data += fourop.vmm2vmem(data[0x060f+128*i:0x60f+128*(i+1)])[0]
                    dx72data += fourop.vmm2vmem(data[0x060f+128*i:0x60f+128*(i+1)])[1]

        if ext == ".b16" and size == 20837:
            if dxcommon.list2string(data[6:18]) == "YAMAHA FB-01":
                for bank in range(2):
                    for i in range(48):
                        fbdata = []
                        for p in range(64):
                            lo = data[0x2029*bank + 0x0821 + 130*i + 2*p] & 15
                            hi = (data[0x2029*bank + 0x0821 + 130*i+ 2*p + 1] & 15) << 4
                            fbdata.append(lo+hi)
                        dx7data += fb01.fb2vmem(fbdata)[0]
                        dx72data += fb01.fb2vmem(fbdata)[1]

        if ext == ".b68" and size==9816:
            if dxcommon.list2string(data[6:16]) == "SIDEMAN DX":
                dx7data += data[0x60f:0x160f]
                for i in range(32):
                    tx7data += data[0x1bd8+84*i:0x1bd8+84*i+64]
        
    # DX7IIFD floppy data
    elif ext[:2] ==".i" and size==16384:
        dx7data = data[0x2000:0x4000]
        for i in range(32):
            dx72data += data[0x1980+26*i:0x1980+26*(i+1)]
            dx72data += data[0x1740+9*i:0x1740+9*(i+1)]
        for i in range(32):
            dx72data += data[0x1cc0+26*i:0x1cc0+26*(i+1)]
            dx72data += data[0x1860+9*i:0x1860+9*(i+1)]

    # DX7 Steinberg Synthworks/Satellit SND
    elif ext==".snd" and size==5216:
        dx7data = data[:4096]

    # FM-Alive DX Manager DXM
    elif ext == ".dxm" and data[:7] == [0xfc, 0, 0, 0, 0x44, 0x58, 0x4d] and size>256:
        if dxcommon.list2string(data[0xfd:0xfd+3]) != 'OP4':
            for i in range(255, size-159):
                if data[i:i+4] == [155, 0, 0, 0]:
                    dx7data += dx7.vced2vmem(data[i+4:i+159])
                    if data[i+159:i+163] == [49, 0, 0, 0]:
                        dx72data += dx7.aced2amem(data[i+163:i+163+49])
                    else:
                        dx72data += dx7.initamem()
                if data[i:i+4] == [128, 0, 0, 0]:
                    dx7data += data[i+4:i+132]
                    if data[i+132:i+136] == [35, 0, 0, 0]:
                        dx72data += data[i+136:i+136+35]
                    else:
                        dx72data += dx7.initamem()

    # Soundlib SLB
    elif ext==".slb" and size>138:
        for i in range(size-139):
            if data[i:i+11] == [0x00, 0x00, 0x08, 0x44, 0x58, 0x37, 0x56, 0x6F, 0x69, 0x63, 0x65]:
                dx7data += data[i+11:i+11+128]


    # Opcode GALAXY Yamaha DX7
    elif size==4170 and data[0x1c:0x1c+10]==[0x59, 0x61, 0x6d, 0x61, 0x68, 0x61, 0x20, 0x44, 0x58, 0x37]:
        if data[-8:]==[0xff, 0xff, 0, 0, 0xff, 0xff, 0, 0]:
            dx7data += data[0x42:0x42+4096]
    
    # C-Lab X-ALyzer XAL
    elif ext==".xal":
        for i in range(32*(size//8000)):
            dx7data += data[252+250*i:380+250*i]
            dx72data += data[380+250*i:415+250*i]
    
    # Amiga DXEditor093
    elif data[:10] == [0x46, 0x4f, 0x52, 0x4d, 0, 0, 0, 0, 0x44, 0x78]:
        for i in range((size-12)//212):
            vced=[]
            for oop in range(6):
                op=5-oop
                vced += data[12+212*i+22*op:12+212*i+22*op+20]
                if data[12+212*i+22*op+20]&128 == 128:
                    vced += [(data[12+212*i+22*op+20]&7)-1]
                else:
                    vced += [7+(data[12+212*i+22*op+20]&7)]
            vced += data[12+212*i+132:12+212*i+132+19]
            vced += data[12+212*i+152:12+212*i+162]
            dx7data += dx7.vced2vmem(vced)
    
    # Marc Bareille SynLib DX-TX DXL
    elif ext==".dxl" and size==4128320:
        for patch in range(32000):
            if data[129*patch:129*patch+118] != [0]*118:
                dx7data += data[129*patch:129*(patch+1)-1]

    # Emagic Sounddiver
    elif ext == ".lib" and data[:4] == [0x4d, 0x52, 0x4f, 0x46]:
        for i in range(size-158):
            if data[i:i+4] == [0x54, 0x4e, 0x45, 0x4c]:
                nmlen=data[i+21]
                if data[i+22+nmlen:i+25+nmlen] == [0x03, 0x81, 0x19]:
                    dx7data += data[i+25+nmlen:i+25+nmlen+118]
                    dx7data += data[i+22:i+22+nmlen] + [32]*(10-nmlen)
                    dx72data += data[i+25+nmlen+118:i+25+nmlen+118+35]
                    i += 127+27
                elif data[i+42+nmlen:i+45+nmlen] == [0x03, 0x83, 0x0f]:
                    dx7data += data[i+45+nmlen:i+45+nmlen+118]
                    dx7data += data[i+22:i+22+nmlen] + [32]*(10-nmlen)
                    dx72data += data[i+45+nmlen+118:i+45+nmlen+118+35]
                    i += 127+47

    # C-Lab PolyFrame
    elif ext == ".pl" and dxcommon.list2string(data[:13]) == "PolyFrame Lib":
        for i in range(len(data)-4):
            if data[i:i+4] == dxcommon.string2list('DX7')+[0]:
                data_start = i+4
                for n in range(i-14, i):
                    if data[n:n+2] == [0x60, 0]:
                        name_start = n+4
                        dx7dat = 118*[0]+10*[32]
                        dx7dat[:118] = data[data_start:data_start+118]
                        dx7dat[118:118+(i-name_start)] = data[name_start:i]
                        dx7data += dx7dat
            elif data[i:i+6] == dxcommon.string2list('DX7II')+[0]:
                data_start = i+6
                for n in range(i-14, i):
                    if data[n:n+2] == [0x60, 0] and (i-n>5):
                        name_start = n+4
                        dx7dat = 118*[0]+10*[32]
                        dx7dat[:118] = data[data_start:data_start+118]
                        dx7dat[118:118+(i-name_start)] = data[name_start:i]
                        dx7data += dx7dat
                    elif data[n:n+2] == [0x60, 0] and (i-n==5):
                        dx72data += data[data_start:data_start+35]
            elif data[i:i+6] == dxcommon.string2list('TX802')+[0]:
                data_start = i+6
                for n in range(i-14, i):
                    if data[n:n+2] == [0x60, 0] and (i-n>5):
                        name_start = n+4
                        dx7dat = 118*[0]+10*[32]
                        dx7dat[:118] = data[data_start:data_start+118]
                        dx7dat[118:118+(i-name_start)] = data[name_start:i]
                        dx7data += dx7dat
                    elif data[n:n+2] == [0x60, 0] and (i-n==5):
                        dx72data += data[data_start:data_start+35]

    # Dr.T FourOp de Luxe
    elif size==4096 and ext==".dxx":
        print "WARNING: 4op to 6op conversion" 
        for i in range(32):
            if sum(data[128*i+118:128*i+128])==0 and sum(data[128*i:128*i+118])!=0:
                dx7data += fourop.vmm2vmem(data[128*i:128+128*i])[0]
                dx72data += fourop.vmm2vmem(data[128*i:128+128*i])[1]

    # X-OR V50
    elif size==24576 and ext==".v5b":
        print "WARNING: 4op to 6op conversion"
        for i in range(100):
            dx7data += fourop.vmm2vmem(data[128*i:128*(i+1)])[0]
            dx72data += fourop.vmm2vmem(data[128*i:128*(i+1)])[1]

    # X-OR TX81Z
    elif size==6538 and ext==".txz":
        print "WARNING: 4op to 6op conversion"
        for i in range(32):
            dx7data += fourop.vmm2vmem(data[128*i:128*(i+1)])[0]
            dx72data += fourop.vmm2vmem(data[128*i:128*(i+1)])[1]

    # V50 internal disk SYN format
    elif size==26624 and ext[:-2]==".i":
        print "WARNING: 4op to 6op conversion" 
        for i in range(100):
            d = data[13824+128*i:13824+128*(i+1)]
            if d[86]<50:
                d[86] += 51
            else:
                d[86] -= 50
            dx7data += fourop.vmm2vmem(d)[0]
            dx72data += fourop.vmm2vmem(d)[1]

    # m.gregory's TX81Z Programmer .tx8
    elif ext=='.tx8' and data[:5] == [0x54, 0x58, 0x38, 0x31, 0x5a]:        
        for i in range((size-9)//173):
            dx7data += fourop.vmm2vmem(data[54+173*i:54+173*i+128])[0]
            dx72data += fourop.vmm2vmem(data[54+173*i:54+173*i+128])[1]

    # Synthworks 4OP
    elif ext=='.bnk' and size==4032 and data[0]==0x0a:
        for i in range(32):
            txdat = fourop.steinb2vmm(data[1+12*i:11+12*i], data[0x180+114*i:0x180+114*(i+1)])
            dx7data += fourop.vmm2vmem(txdat)[0]
            dx72data += fourop.vmm2vmem(txdat)[1]
    elif ext=='.snd' and size==126 and data[0]==0x0a:
        txdat = fourop.steinb2vmm(data[1:11], data[12:])
        dx7data += fourop.vmm2vmem(txdat)[0]
        dx72data += fourop.vmm2vmem(txdat)[1]

    # Cages Artist FB01
    elif size==8768 and ext==".fb1":
        print "WARNING: FB01 to 6op conversion"
        for bnk in range(2):
            for i in range(48):
                dx7data += fb01.fb2vmem(data[32+3104*bnk+64*i:32+3104*bnk+64*(i+1)])[0]
                dx72data += fb01.fb2vmem(data[32+3104*bnk+64*i:32+3104*bnk+64*(i+1)])[1]
    
    # Dr.T FB01 48 * 64 bytes
    elif size==3072 and ext==".fb1":
        print "WARNING: FB01 to 4op TX/DX conversion" 
        for i in range(48):
            dx7data += fb01.fb2vmem(data[64*i:64*(i+1)])[0]
            dx72data += fb01.fb2vmem(data[64*i:64*(i+1)])[1]

    # Frederic Meslin FB01 editor
    elif size==354 and ext==".fbv":
        pass
        #print "**WARNING: FB01 to 4op TX/DX conversion" 
        #dx7data += fb01.fb2vmem(data[0x107:0x107+64])[0]
        #dx72data += fb01.fb2vmem(data[0x107:0x107+64])[1]
    
    # Synthworks FB01
    elif size==2450 and ext==".fbd":
        print "WARNING: FB01 to 4op TX/DX conversion"
        for i in range(49):
            fb = 64*[0]
            fb[:7] = data[1+8*i:8+8*i]
            fb[8:48] = data[0x188+42*i:0x1b0+42*i]
            fb[58:60] = data[0x1b0+42*i:0x1b2+42*i]
            dx7data += fb01.fb2vmem(fb)[0]
            dx72data += fb01.fb2vmem(fb)[1]

    # VOPM .OPM
    elif ext==".opm":
        dx7data = []
        dx72data = []
        tx7data = []
        print "WARNING: experimental VOPM (.opm) to 6-OP conversion"
        data = vopm.file2data(infile)
        for i in range(len(data)//76):
            fb = vopm.vopm2fb01(data[76*i:76*(i+1)])
            dx = fb01.fb2vmem(fb)[0]
            dx[118:128] = data[76*i:76*i+10]
            dx7data += dx
            dx72data += fb01.fb2vmem(fb)[1]
 
    # geminisolstice "Michael Jackson" collection Yamaha DX7, TX7 and TX816
    elif size>192:
        sizecheck=(data[0]<<8)+data[1]
        if (size==2+(192*sizecheck)) or (size==32+(192*sizecheck)):
            for i in range(sizecheck):
                dx7data += data[2+192*i:2+192*i+128]
                tx7data += data[2+192*i+128:2+192*i+192]
        elif (size==2+(128*sizecheck)) or (size==32+(128*sizecheck)):
            for i in range(sizecheck):
                dx7data += data[2+128*i:2+128*(i+1)]
    
    # single VCED 
    elif (size%155)==0:
        if (size%128)!=0:
            for i in range(size//155):
                dx7data += dx7.vced2vmem(data[155*i:155*(i+1)])
   
    # Ignore many other SysEx
    if dx7data == []:
        for i in data[:64]:
            if i > 127:
                dx7data += dx7.initvmem()
                break

    # If everything fails: The worst thing that can happen is
    # creating a DX7 compatible but completely useless patchbank :-)
    # Should work with e.g. headerless 4096 bytes DX7 files
    if dx7data == []:
        dx7data += data[:128*(len(data)//128)]
    if dx72data == []:
        if len(tx7data)//64 == len(dx7data)//128:
            for i in range(len(dx7data)//128):
                dx72data += tx7.pmem2amem(tx7data[64*i:64*(i+1)])
        else:
            dx72data = dx7.initamem()*(len(dx7data)//128)
    if tx7data == []:
        if len(dx72data)//35 == len(dx7data)//128:
            for i in range(len(dx7data)//128):
                tx7data += tx7.amem2pmem(dx72data[35*i:35*(i+1)])
        else:
            tx7data = tx7.initpmem()*(len(dx7data)//128)


    # Safety garbage cleanup
    dx7dat = []
    dx72dat = []
    tx7dat = []
    initcount = 0
    for i in range(len(dx7data)//128):        
        if dxcommon.list2string(dx7data[128*i+118:128*i+128]) != "INIT VOICE":
            if dxcommon.list2string(dx7data[128*i+118:128*i+128]).strip() != "":
                dx7dat += dx7.cleanvmem(dx7data[128*i:128*(i+1)])
                dx72dat += dx7.cleanamem(dx72data[35*i:35*(i+1)])
                tx7dat += tx7.cleanpmem(tx7data[64*i:64*(i+1)])
        else:
            initcount += 1

    # restore INITVOICE
    dx7dat += dx7.initvmem() * initcount
    dx72dat += dx7.initamem() * initcount
    tx7dat += tx7.initpmem() * initcount

    return dx7dat, dx72dat, tx7dat, channel

def pre_write(dx7data, dx72data, tx7data, dx72, TX7):
    # AMS AMEM/VMEM correction and conversion
    ams=(0, 1, 2, 3, 3, 3, 3, 3)
    # ams=(0, 1, 1, 2, 2, 2, 3, 3)
    for p in range(len(dx7data)//128):
        for op in range(6):
            ams1=dx7data[128*p+17*(5-op)+13]&3
            if op in (0, 2, 4):
                ams2 = (dx72data[35*p+1+((5-op)//2)]//8)&7
            if op in (1, 3, 5):
                ams2 = dx72data[35*p+1+((5-op)//2)]&7
            if ams2 == 0:
                ams2 = ams1
            ams1 = ams[ams2]
            dx7data[128*p+17*(5-op)+13] = ams1 + (dx7data[128*p+17*(5-op)+13]&0b11100)
            if op in (0, 2, 4):
                dx72data[35*p+1+((5-op)//2)] = (dx72data[35*p+1+((5-op)//2)]&7) + 8*ams2
            if op in (1, 3, 5):
                dx72data[35*p+1+((5-op)//2)] = (dx72data[35*p+1+((5-op)//2)]&0b111000) + ams2
    # Change Key Scaling Mode from Fractional to Normal
    for p in range(len(dx72data)//35):
        dx72data[35*p]=0

    return dx7data, dx72data, tx7data

def write(outfile, dx7data, dx72data, tx7data, dx72, TX7, channel, nosplit, mid_out=MID_OUT):
    dx7data, dx72data, tx7data = pre_write(dx7data, dx72data, tx7data, dx72, TX7)
    ext = os.path.splitext(outfile)[1]
    basename = os.path.split(outfile)[1]
    basename = os.path.splitext(basename)[0]
    dirname = os.path.dirname(outfile)
    mid = False
    if ext.lower() in (".mid", ".midi"):
        mid = True
    if ext.lower() == ".dx7":
        dx72 = False
    if dxcommon.ENABLE_MIDI:
        if os.path.basename(outfile) == "MIDI":
            nosplit = True
 
    # PEG range DX7II AMEM to DX7 VMEM conversion:
    if dx72==False:
        for p in range(len(dx7data)//128):
            for i in range(106, 110):
                pegr=dx72data[35*p+4]&3
                peg=dx7data[128*p+i]
                # dx7data[128*p+i]=int((peg-50)*(50, 32, 18, 10)[pegr]/50.0) + 50
                dx7data[128*p+i]=int((peg-50)*(50, 32, 16, 8)[pegr]/50.0) + 50

    if len(dx7data)>128:
        #Fill unused patch locations in 32-voice banks with INIT VOICE
        while len(dx7data)%4096 != 0:
            dx7data += dx7.initvmem()
            dx72data += dx7.initamem()
            tx7data += tx7.initpmem()
        patchcount=len(dx7data)//128
    elif len(dx7data)<128:
        dx7data, dx72data, tx7data = dx7.initvmem(), dx7.initamem(), tx7.initpmem()
        patchcount=1
    else:
        patchcount=1
    dat=[]
    message = ''
    bankcount=max(1, len(dx7data)//4096)
    bankdigits=len(str(bankcount+1))
    for bank in range(bankcount):
        if len(dx7data)//4096>1 and nosplit==False:
            outfile=os.path.join(dirname, basename+"("+str(bank+1).zfill(bankdigits)+")"+ext)
        if nosplit==False:
            dat=[]

        if ext.lower()==".dx7":
            if patchcount==1:
                dat = dx7data
            else:
                dat += dx7data[4096*bank:4096*(bank+1)]

        elif ext.lower()==".txt":
            if patchcount==1:
                dat = dx7.vmem2txt(dx7data)
                if dx72:
                    dat += dx7.amem2txt(dx72data)
                if TX7:
                    dat += tx7.pmem2txt(tx7data)
            else:
                t = ''
                for i in range(32):
                    t += "{:03}: {}\n".format(i+1+bank*32, dx7.voicename(dx7data[4096*bank+128*i:4096*bank+128*(i+1)]))
                t += "\n"

                for i in t:
                    dat.append(ord(i))
        
        else: # syx and mid files
            if patchcount==1: #save as VCED sysex
                if dx72:
                    dat += [0xf0, 0x43, channel, 0x05, 0x00, 0x31]
                    dat += dx7.amem2aced(dx72data)
                    dat += [dxcommon.checksum(dx7.amem2aced(dx72data)), 0xf7]
                dat += [0xf0, 0x43, channel, 0x00, 0x01, 0x1b]
                dat += dx7.vmem2vced(dx7data)
                dat += [dxcommon.checksum(dx7.vmem2vced(dx7data)), 0xf7]
                if TX7:
                    dat += [0xf0, 0x43, channel, 0x01, 0x00, 0x5e]
                    dat += tx7.pmem2pced(tx7data)
                    dat += [dxcommon.checksum(tx7.pmem2pced(tx7data)), 0xf7]
                
            else: #save as VMEM sysex
                if nosplit==True:
                    if patchcount > 32:
                        dat += [0xf0, 0x43, 0x10+channel, 0x19, 0x4d, bank%128, 0xf7]
                if dx72:
                    dat += [0xf0, 0x43, channel, 0x06, 0x08, 0x60]
                    dat += dx72data[1120*bank:1120*(bank+1)]
                    dat += [dxcommon.checksum(dx72data[1120*bank:1120*(bank+1)]), 0xf7]
                dat += [0xf0, 0x43, channel, 0x09, 0x20, 0x00]
                dat += dx7data[4096*bank:4096*(bank+1)]
                dat += [dxcommon.checksum(dx7data[4096*bank:4096*(bank+1)]), 0xf7]
                if TX7:
                    dat += [0xf0, 0x43, channel, 0x02, 0x20, 0x00]
                    dat += tx7data[2048*bank:2048*(bank+1)]*2
                    dat += [dxcommon.checksum(tx7data[2048*bank:2048*(bank+1)]*2), 0xf7]

        if nosplit==False:
            if mid:
                dat = syxmidi.syx2mid(dat)
            with open(outfile, 'wb') as f:
                array.array('B', dat).tofile(f)
            print "{} Patch(es) written to: {}".format(min(32, len(dx7data)//128), outfile)

    if nosplit==True:
        if mid:
            dat = syxmidi.syx2mid(dat)
        with open(outfile, 'wb') as f:
            array.array('B', dat).tofile(f)
        print "{} Patch(es) written to: {}".format(len(dx7data)//128, outfile)

    if dxcommon.ENABLE_MIDI:
        if os.path.basename(outfile) == "MIDI":
            dxcommon.data2midi(dat, mid_out)

    return "Ready. {} Patch(es) written to output file(s)".format(len(dx7data)//128)

def dx2list(dx7data, dx72data, tx7data, dx72, TX7):
    dxlist = []
    for i in range(len(dx7data)//128):
        dx7dat = dx7data[128*i:128*(i+1)]
        dx72dat = dx72data[35*i:35*(i+1)]
        tx7dat = tx7data[64*i:64*(i+1)]

        a = dx7dat[118:128] + dx7dat[:118]
        a += dx72dat
        a += tx7dat

        dxlist.append(a)
    return dxlist

def list2dx(data):
    dx7data, dx72data, tx7data = [], [], []
    for i in data:
        dx7data += i[10:128] + i[:10]
        dx72data += i[128:128+35]
        tx7data += i[128+35:128+35+64]
    return dx7data, dx72data, tx7data

def dxsort(dx7data, dx72data, tx7data, dx72, TX7):
    if (len(dx7data)//128) > 1:
        print "Sorting patches by name ..."
        l = dx2list(dx7data, dx72data, tx7data, dx72, TX7)
        L = []
        for n in range(len(l)):
            if dxcommon.list2string(l[n][:10]) != "INIT VOICE":
                L.append(l[n])
        L.sort()
        dx7data, dx72data, tx7data = list2dx(L)
    return dx7data, dx72data, tx7data

def dxnodupes(dx7data, dx72data, tx7data, dx72, TX7, nodupes2):
    if (len(dx7data)//128) > 1:
        print "Searching and removing duplicates ..."
        L = dx2list(dx7data, dx72data, tx7data, dx72, TX7)
        l=[L[0]]
        if nodupes2 == True:
            for i in range(len(L)):
                dup = False
                for j in range(len(l)):
                    if l[j][10:] == L[i][10:]:
                        dup=True
                        break
                if dup == False:
                    l.append(L[i])
        else:
            for i in L:
                if not i in l:
                    l.append(i)
        dx7data, dx72data, tx7data = list2dx(l)
    return dx7data, dx72data, tx7data


def dxrandom(dx7data, deviation):
    print 'Randomizing data ...'
    import random
    n = len(dx7data) // 128

    if deviation == 0:
        # thanks to Renata Dokmanovic
        newdata = [] + dx7data
        for patch in range(n):
            for param in range(128):
                newdata[128 * patch + param] = dx7data[128 * random.randrange(n) + param]
        return newdata

    newdata = []
    pmax = [0] * 128
    pmin = [127] * 128
    deviation = min(deviation, 300)
    dev = deviation % 100

    for p in range(128):
        for i in range(n):
            if dx7data[128*i + p] > pmax[p]:
                pmax[p] = dx7data[128*i + p]
            if dx7data[128*i + p] < pmin[p]:
                pmin[p] = dx7data[128*i + p]

    for i in range(n):
        for p in range(128):
            pmax[p] = dx7data[128*i + p] + int(round((pmax[p] - dx7data[128*i + p]) * (dev/100.)))
            pmin[p] = dx7data[128*i + p] - int(round((dx7data[128*i + p] - pmin[p]) * (dev/100.)))
            newdata.append(random.randint(pmin[p], pmax[p]))
        
        alg = dx7data[128*i + 110]
        algnew = newdata[128*i + 110]
        if deviation < 201: # don't randomize algo
            newdata[128*i + 110] = alg
            algnew = alg
        
        # keep carrier frequencies playable
        newdata[128*i + 106:128*i + 110] = dx7data[128*i + 106:128*i + 110] # PEG levels
        newdata[128*i + 117] = dx7data[128*i + 117] # Transpose
        for op in range(6):
            if dx7.carrier(algnew, op) and dx7.carrier(alg, op):
                for p in (12, 15, 16):
                    newdata[128*i + 17*(5 - op) + p] = dx7data[128*i + 17*(5 - op) + p]
            elif dx7.carrier(algnew, op): # use OP1 frequency
                for p in (12, 15, 16):
                    newdata[128*i + 17*(5 - op) + p] = dx7data[128*i + 17*(5 - 0) + p]
        
        if deviation < 101: # don't randomize modulator frequencies either
            for op in range(6):
                if not dx7.carrier(alg, op):
                    for p in (12, 15, 16):
                        newdata[128*i + 17*(5 - op) + p] = dx7data[128*i + 17*(5 - op) + p]

        # auto create nice patchnames, 
        #reflecting the deviation value
        if newdata[128*i:128*i + 118] == dx7data[128*i:128*i + 118]:
            newdata[128*i + 118:128*i + 128] == dx7data[128*i + 118:128*i + 128]
        else:
            for p in range(118, 128):
                if (p-118) < (10 - dev//10):
                    newdata[128*i + p] = dx7data[128*i + p]
                else:
                    newdata[128*i + p] = random.randrange(65, 91)

    return newdata

def dxfind(dx7data, dx72data, tx7data, keyword):
    dx7dat = []
    dx72dat = []
    tx7dat = []
    for i in range(len(dx7data)//128):
        s = ''
        for k in range(128*i+118, 128*i+128):
            s += chr(dx7data[k])
        if keyword.lower() in s.lower():
            dx7dat += dx7data[128*i:128*i+128]
            dx72dat += dx72data[35*i:35*i+35]
            tx7dat += tx7data[64*i:64*i+64]
    return dx7dat, dx72dat, tx7dat


