# read .opm files from VOPM OPM synth emulator
# and convert to FB01 format
from . import dxcommon
import os

def file2data(opmfile):
    with open(opmfile, 'r') as f:
        lines = f.readlines()
   
    data=[]
    p=-1
    for i in range(len(lines)):
        if lines[i][0] == '@':
            p += 1
            nm = lines[i][3+len(str(p)):].strip()
            name = 20*[32] # This is probably more than we really need
            for n in range(min(20, len(nm))):
                name[n] = ord(nm[n])

            for initvoice in ("init      ", "Null      ", "no Name   "):
                n = dxcommon.list2string(name[:10])
                if n.lower() == initvoice.lower():
                    name[:10] = [73, 78, 73, 84, 32, 86, 79, 73, 67, 69] # "INIT VOICE"
                    break
            
            if dxcommon.list2string(name)[:10].lower() == "instrument":
                x = "{:04}".format(int(dxcommon.list2string(name[11:15])))
                newname = os.path.basename(opmfile)[:-4]
                if len(newname)<6:
                    newname += "_"*(6-len(newname))
                newname = newname[:6]+"{}".format(x)
                name[:10] = dxcommon.string2list(newname)

            lfo = lines[i+1][4:].split()
            ch = lines[i+2][4:].split()
            m1 = lines[i+3][4:].split()
            c1 = lines[i+4][4:].split()
            m2 = lines[i+5][4:].split()
            c2 = lines[i+6][4:].split()
            
            # stringnumbers to intnumbers
            for i in range(len(lfo)):
                lfo[i] = int(lfo[i])
            for i in range(len(ch)):
                ch[i] = int(ch[i])
            for i in range(len(m1)):
                m1[i] = int(m1[i])
            for i in range(len(c1)):
                c1[i] = int(c1[i])
            for i in range(len(m2)):
                m2[i] = int(m2[i])
            for i in range(len(c2)):
                c2[i] = int(c2[i])
            
            data += (name+lfo+ch+m1+c1+m2+c2)
    return data

def vopm2fb01(vopm):
    fb = [0]*64
    fb[0:7] = vopm[0:7] #NAME
    fb[8] = vopm[20]&255 # LFO speed
    fb[9] = vopm[21]&127 # AMD
    fb[10] = vopm[22]&127 #PMD
    fb[11] = vopm[30]&120 #Operator enable
    fb[12] = ((vopm[26]&7)<<3) + (vopm[27]&7) # FBL, ALG
    fb[13] = (vopm[28]&3) + ((vopm[29]&3)<<4) # AMS, PMS
    fb[14] = (vopm[23]&3)<<5
    fb[15] = 0 #Transpose
    for op in range(4):
        if ((vopm[30]>>(op+3))&1) == 0:
            fb[0+16+8*op] = 127
        else:
            fb[0+16+8*op] = vopm[37+11*op]&127 #TL
        fb[1+16+8*op] = 0   # KStype, Velocity sensitivity
        fb[2+16+8*op] = 0   # Level Scaling, TL adjust
        if op in (1, 3):
            pass #fb[2+16+8*op] = 15
        fb[3+16+8*op] = (vopm[39+11*op]&15) + ((vopm[40+11*op]&7)<<4) #MUL, DT1
        fb[4+16+8*op] = (vopm[32+11*op]&31) + ((vopm[38+11*op]&3)<<6) #AR, RS
        fb[5+16+8*op] = (vopm[42+11*op]&128) + (vopm[33+11*op]&31)    #AMS-EN, D1R 
        fb[6+16+8*op] = (vopm[34+11*op]&31) + ((vopm[41+11*op]&3)<<6)  #D2R, DT2
        fb[7+16+8*op] = (vopm[35+11*op]&15) + ((vopm[36+11*op]&15)<<4)   #RR, D1L
    return fb



